import os,csv,re, time, random
import cv2
import numpy as np
import pandas as pd
import torch
import scanpy as sc
from . util import *
from . contour_util import *
from . calculate_dis import *
from . TESLA import TESLA


def annotation(img, 
			sudo_adata, 
			genes, 
			resize_factor, 
			binary, 
			res=50, 
			num_required=1, 
			pooling="min", 
			rseed=100, 
			tseed=100, 
			nseed=100, 
			nChannel=100, 
			threshold=1000, 
			radius=3, 
			minLabels=30, 
			train_refine=True, 
			plot_intermedium=False,
			target_size="small"):
	#-------------------------------Image band-------------------------------------------------#
	if target_size=="small":
		target_ratio=1/2
	elif target_size=="large":
		target_ratio=1/3
	else:
		print("target_size not valid, Please specify (small / large).Use default small")
		target_ratio=1/2
	print("Computing image band...")
	resize_width=int(img.shape[1]*resize_factor)
	resize_height=int(img.shape[0]*resize_factor)
	img_resized=(img * np.dstack([(binary!=0)]*3)).astype(np.uint8)
	img_resized = cv2.resize(img_resized, (resize_width, resize_height), interpolation = cv2.INTER_AREA)
	gray_resized = cv2.cvtColor(img_resized,cv2.COLOR_BGR2GRAY)
	gray_resized=gray_resized.reshape(list(gray_resized.shape)+[1])
	binary_resized=cv2.resize(binary, (resize_width, resize_height), interpolation = cv2.INTER_AREA)
	img_band1=(gray_resized-np.min(gray_resized))/(np.max(gray_resized)-np.min(gray_resized))
	#-------------------------------Gene band-------------------------------------------------#
	print("Computing gene band...")
	genes=list(set([i for i in genes if i in sudo_adata.var.index ]))
	assert num_required<=len(genes)
	tmp=sudo_adata.X[:,sudo_adata.var.index.isin(genes)]
	tmp=(tmp-np.min(tmp, 0))/(np.max(tmp, 0)-np.min(tmp, 0))
	tmp = np.partition(tmp, -num_required, axis=1)[:,-num_required:] #Select top num_required
	if pooling=="mean":
		sudo_adata.obs["meta"]=np.mean(tmp, axis=1)
	elif pooling=="min":
		sudo_adata.obs["meta"]=np.min(tmp, axis=1)
	else:
		print("Error! Pooling logic not understood.")
	gene_img=np.zeros(list(img.shape[0:2])+[1])
	for _, row in sudo_adata.obs.iterrows():
		x=row["x"]
		y=row["y"]
		exp=row["meta"]
		gene_img[int(x-res/2):int(x+res/2), int(y-res/2):int(y+res/2),0]=exp
	gene_img[binary==0]=0
	gene_band1=gene_img[:,:,0]
	gene_band1=cv2.resize(gene_band1, (resize_width, resize_height), interpolation = cv2.INTER_AREA)
	gene_band1=(gene_band1-np.min(gene_band1))/(np.max(gene_band1)-np.min(gene_band1))
	gene_band1=gene_band1.reshape(list(gene_band1.shape)+[1])
	#-------------------------------TESLA-------------------------------------------------#
	print("Running TESLA...")
	assert np.max(img_band1)==1 and np.min(img_band1)==0 and np.max(gene_band1)==1 and np.min(gene_band1)==0
	data=np.concatenate((img_band1, gene_band1), axis=2)
	random.seed(rseed)
	torch.manual_seed(tseed)
	np.random.seed(nseed)
	tesla=TESLA()
	tesla.train(input=data, use_cuda=False, train_refine=train_refine, radius=radius, nChannel=nChannel, lr=0.1, 
		minLabels=minLabels, maxIter=30, stepsize_sim=1, stepsize_con=5, threshold=threshold, plot_intermedium=plot_intermedium, plot_dir=None)
	prob, pred = tesla.predict(data)
	pred_refined=refine_clusters(pred=pred, resize_height=resize_height, resize_width=resize_width, threshold=threshold, radius=radius)
	nLabels=len(np.unique(pred))
	mainLabels=len(np.unique(pred_refined))
	#-----------------------------------Find target cluster---------------------------#
	print("Finding target clusters...")	
	marker=gene_band1.flatten()
	clusters=pred_refined.flatten()
	c_m={} #cluster_marker expression
	for i in np.unique(clusters):
		c_m[i]=np.mean(marker[clusters==i])
	c_m = sorted(c_m.items(), key=lambda x: x[1], reverse=True)
	target_clusters=list(filter(lambda x: (x[1] > c_m[0][1]*target_ratio), c_m))
	target_clusters=[x[0] for x in target_clusters]
	print("c_m:\n", c_m, "\n", "Target clusters:\n", target_clusters, sep = '')
	return pred_refined, target_clusters, c_m

def visualize_annotation(pred_refined, 
				target_clusters, 
				c_m, 
				img,
				binary, 
				resize_factor,
				cnt_color=clr.LinearSegmentedColormap.from_list('red', ["#EAE7CC", '#BA0000'], N=256)):
	resize_width=int(img.shape[1]*resize_factor)
	resize_height=int(img.shape[0]*resize_factor)
	binary_resized=cv2.resize(binary, (resize_width, resize_height), interpolation = cv2.INTER_AREA)
	background = cv2.resize(img, (resize_width, resize_height), interpolation = cv2.INTER_AREA)
	ret_img=(background.copy()).astype(np.uint8)
	alpha=0.8
	#Whiten
	white_ratio=0.5
	ret_img=ret_img*(1-white_ratio)+np.array([255, 255, 255])*(white_ratio)
	for i in range(len(target_clusters)):
		color=((np.array(cnt_color(int(c_m[i][1]/c_m[0][1]*255)))[0:3])*255).astype(int)[::-1]
		target_img=(1*(pred_refined==target_clusters[i])).reshape(resize_height, resize_width)
		target_img[binary_resized==0]=0
		ret_img[target_img!=0]=ret_img[target_img!=0]*(1-alpha)+np.array(color)*(alpha)
	return ret_img






__version__ = '1.2.0'
from . TESLA import *
from . util import *
from . contour_util import *
from . calculate_dis import *
from . imputation import *
from . annotation import *
from . contour_util import *



import scanpy as sc
import pandas as pd
import numpy as np
import scipy
import os
from anndata import AnnData,read_csv,read_text,read_mtx
from scipy.sparse import issparse

def prefilter_cells(adata,min_counts=None,max_counts=None,min_genes=200,max_genes=None):
    if min_genes is None and min_counts is None and max_genes is None and max_counts is None:
        raise ValueError('Provide one of min_counts, min_genes, max_counts or max_genes.')
    id_tmp=np.asarray([True]*adata.shape[0],dtype=bool)
    id_tmp=np.logical_and(id_tmp,sc.pp.filter_cells(adata.X,min_genes=min_genes)[0]) if min_genes is not None  else id_tmp
    id_tmp=np.logical_and(id_tmp,sc.pp.filter_cells(adata.X,max_genes=max_genes)[0]) if max_genes is not None  else id_tmp
    id_tmp=np.logical_and(id_tmp,sc.pp.filter_cells(adata.X,min_counts=min_counts)[0]) if min_counts is not None  else id_tmp
    id_tmp=np.logical_and(id_tmp,sc.pp.filter_cells(adata.X,max_counts=max_counts)[0]) if max_counts is not None  else id_tmp
    adata._inplace_subset_obs(id_tmp)
    adata.raw=sc.pp.log1p(adata,copy=True) #check the rowname 
    print("the var_names of adata.raw: adata.raw.var_names.is_unique=:",adata.raw.var_names.is_unique)
   

def prefilter_genes(adata,min_counts=None,max_counts=None,min_cells=10,max_cells=None):
    if min_cells is None and min_counts is None and max_cells is None and max_counts is None:
        raise ValueError('Provide one of min_counts, min_genes, max_counts or max_genes.')
    id_tmp=np.asarray([True]*adata.shape[1],dtype=bool)
    id_tmp=np.logical_and(id_tmp,sc.pp.filter_genes(adata.X,min_cells=min_cells)[0]) if min_cells is not None  else id_tmp
    id_tmp=np.logical_and(id_tmp,sc.pp.filter_genes(adata.X,max_cells=max_cells)[0]) if max_cells is not None  else id_tmp
    id_tmp=np.logical_and(id_tmp,sc.pp.filter_genes(adata.X,min_counts=min_counts)[0]) if min_counts is not None  else id_tmp
    id_tmp=np.logical_and(id_tmp,sc.pp.filter_genes(adata.X,max_counts=max_counts)[0]) if max_counts is not None  else id_tmp
    adata._inplace_subset_var(id_tmp)


def prefilter_specialgenes(adata,Gene1Pattern="ERCC",Gene2Pattern="MT-"):
    id_tmp1=np.asarray([not str(name).startswith(Gene1Pattern) for name in adata.var_names],dtype=bool)
    id_tmp2=np.asarray([not str(name).startswith(Gene2Pattern) for name in adata.var_names],dtype=bool)
    id_tmp=np.logical_and(id_tmp1,id_tmp2)
    adata._inplace_subset_var(id_tmp)

def relative_func(expres):
    #expres: an array counts expression for a gene
    maxd = np.max(expres) - np.min(expres)
    min_exp=np.min(expres)
    rexpr = (expres - min_exp)/maxd
    return rexpr

def plot_relative_exp(input_adata, gene, x_name, y_name,color,use_raw=False, spot_size=200000):
    adata=input_adata.copy()
    if use_raw:
        X=adata.raw.X
    else:
        X=adata.X
    if issparse(X):
        X=pd.DataFrame(X.A)
    else:
        X=pd.DataFrame(X)
    X.index=adata.obs.index
    X.columns=adata.var.index
    rexpr=relative_func(X.loc[:,gene])
    adata.obs["rexpr"]=rexpr
    fig=sc.pl.scatter(adata,x=x_name,y=y_name,color="rexpr",title=gene+"_rexpr",color_map=color,show=False,size=spot_size/adata.shape[0])
    return fig

def plot_log_exp(input_adata, gene, x_name, y_name,color,use_raw=False):
    adata=input_adata.copy()
    if use_raw:
        X=adata.X
    else:
        X=adata.raw.X
    if issparse(X):
        X=pd.DataFrame(X.A)
    else:
        X=pd.DataFrame(X)
    X.index=adata.obs.index
    X.columns=adata.var.index
    adata.obs["log"]=np.log((X.loc[:,gene]+1).tolist())
    fig=sc.pl.scatter(adata,x=x_name,y=y_name,color="log",title=gene+"_log",color_map=color,show=False,size=200000/adata.shape[0])
    return fig

def refine_clusters(pred, resize_height, resize_width, threshold, radius):
    pixel_num=pd.Series(pred).value_counts()
    clusters=pixel_num.index.tolist()
    reorder_map={}
    for i in range(pixel_num.shape[0]):
        reorder_map[clusters[i]]=i
    pred_reordered=pd.Series(pred).replace(reorder_map).to_numpy()
    pixel_num=pd.Series(pred_reordered).value_counts()
    # Number of clusters
    nLabels = len(np.unique(pred_reordered))
    # Number of main clusters
    mainLabels=(pd.Series(pred_reordered).value_counts()>=threshold).sum()
    #------------- Refine clusters ---------------------
    main_clusters=pixel_num.index[pixel_num>=threshold].tolist()
    minor_clusters=pixel_num.index[pixel_num<threshold].tolist()
    pred_reordered_img = pred_reordered.reshape( (resize_height, resize_width))
    max_x, max_y=resize_width, resize_height
    replace_map={}
    for i in minor_clusters:
        nbs=[]
        xy=np.where(pred_reordered_img==i)
        for j in range(len(xy[0])):
            x, y=xy[0][j], xy[1][j]
            nbs=nbs+pred_reordered_img[max(0,x-radius):min(max_x,x+radius+1),max(0,y-radius):min(max_y,y+radius+1)].flatten().tolist()
            nbs_num=pd.Series(nbs).value_counts()
            if sum(nbs_num.index.isin(main_clusters))>0:
                replace_map[i]=nbs_num.index[ nbs_num.index.isin(main_clusters) ][ 0 ]
    pred_refined=pd.Series(pred_reordered).replace(replace_map).to_numpy()
    return pred_refined



import os,csv,re, time
import cv2
import pandas as pd
import numpy as np
from . util import *
from . contour_util import *
from . calculate_dis import *

def imputation(img, raw, cnt, genes, shape="None", res=50, s=1, k=2, num_nbs=10):
	binary=np.zeros((img.shape[0:2]), dtype=np.uint8)
	cv2.drawContours(binary, [cnt], -1, (1), thickness=-1)
	#Enlarged filter
	cnt_enlarged = scale_contour(cnt, 1.05)
	binary_enlarged = np.zeros(img.shape[0:2])
	cv2.drawContours(binary_enlarged, [cnt_enlarged], -1, (1), thickness=-1)
	x_max, y_max=img.shape[0], img.shape[1]
	x_list=list(range(int(res), x_max, int(res)))
	y_list=list(range(int(res), y_max, int(res)))
	x=np.repeat(x_list,len(y_list)).tolist()
	y=y_list*len(x_list)
	sudo=pd.DataFrame({"x":x, "y": y})
	sudo=sudo[sudo.index.isin([i for i in sudo.index if (binary_enlarged[sudo.x[i], sudo.y[i]]!=0)])]
	b=res
	sudo["color"]=extract_color(x_pixel=sudo.x.tolist(), y_pixel=sudo.y.tolist(), image=img, beta=b, RGB=True)
	z_scale=np.max([np.std(sudo.x), np.std(sudo.y)])*s
	sudo["z"]=(sudo["color"]-np.mean(sudo["color"]))/np.std(sudo["color"])*z_scale
	sudo=sudo.reset_index(drop=True)
	#------------------------------------Known points---------------------------------#
	known_adata=raw[:, raw.var.index.isin(genes)]
	known_adata.obs["x"]=known_adata.obs["pixel_x"]
	known_adata.obs["y"]=known_adata.obs["pixel_y"]
	known_adata.obs["color"]=extract_color(x_pixel=known_adata.obs["pixel_x"].astype(int).tolist(), y_pixel=known_adata.obs["pixel_y"].astype(int).tolist(), image=img, beta=b, RGB=False)
	known_adata.obs["z"]=(known_adata.obs["color"]-np.mean(known_adata.obs["color"]))/np.std(known_adata.obs["color"])*z_scale
	#-----------------------Distance matrix between sudo and known points-------------#
	start_time = time.time()
	dis=np.zeros((sudo.shape[0],known_adata.shape[0]))
	x_sudo, y_sudo, z_sudo=sudo["x"].values, sudo["y"].values, sudo["z"].values
	x_known, y_known, z_known=known_adata.obs["x"].values, known_adata.obs["y"].values, known_adata.obs["z"].values
	print("Total number of sudo points: ", sudo.shape[0])
	for i in range(sudo.shape[0]):
		if i%1000==0:print("Calculating spot", i)
		cord1=np.array([x_sudo[i], y_sudo[i], z_sudo[i]])
		for j in range(known_adata.shape[0]):
			cord2=np.array([x_known[j], y_known[j], z_known[j]])
			dis[i][j]=distance(cord1, cord2)
	print("--- %s seconds ---" % (time.time() - start_time))
	dis=pd.DataFrame(dis, index=sudo.index, columns=known_adata.obs.index)
	#-------------------------Fill gene expression using nbs---------------------------#
	sudo_adata=AnnData(np.zeros((sudo.shape[0], len(genes))))
	sudo_adata.obs=sudo
	sudo_adata.var=known_adata.var
	#Impute using all spots, weighted
	for i in range(sudo_adata.shape[0]):
		if i%1000==0:print("Imputing spot", i)
		index=sudo_adata.obs.index[i]
		dis_tmp=dis.loc[index, :].sort_values()
		nbs=dis_tmp[0:num_nbs]
		dis_tmp=(nbs.to_numpy()+0.1)/np.min(nbs.to_numpy()+0.1) #avoid 0 distance
		if isinstance(k, int):
			weights=((1/(dis_tmp**k))/((1/(dis_tmp**k)).sum()))
		else:
			weights=np.exp(-dis_tmp)/np.sum(np.exp(-dis_tmp))
		row_index=[known_adata.obs.index.get_loc(i) for i in nbs.index]
		sudo_adata.X[i, :]=np.dot(weights, known_adata.X[row_index,:])
	return sudo_adata




#from __future__ import print_function
import torch
import torch.nn as nn
import torch.nn.functional as F
import torch.optim as optim
import numpy as np
import torch.nn.init

class Spatial_CNN(nn.Module):
	def __init__(self,input_dim, nChannel=100, nConv=2, kernel_size_list=[3,3],stride_list=[1,1], padding_list=[1,1]):
		super(Spatial_CNN, self).__init__()
		assert nConv==len(kernel_size_list)==len(stride_list)==len(padding_list)
		self.nChannel=nChannel
		self.nConv=nConv
		self.conv1 = nn.Conv2d(input_dim, self.nChannel, kernel_size=kernel_size_list[0], stride=stride_list[0], padding=padding_list[0] )
		self.bn1 = nn.BatchNorm2d(self.nChannel)
		self.conv2 = nn.ModuleList()
		self.bn2 = nn.ModuleList()
		for i in range(self.nConv-1):
			self.conv2.append( nn.Conv2d(self.nChannel, self.nChannel, kernel_size=kernel_size_list[i+1], stride=stride_list[i+1], padding=padding_list[i+1] ) )
			self.bn2.append( nn.BatchNorm2d(self.nChannel) )
		self.conv3 = nn.Conv2d(nChannel, nChannel, kernel_size=1, stride=1, padding=0 )
		self.bn3 = nn.BatchNorm2d(self.nChannel)
	def forward(self, x):
		x = self.conv1(x)
		x = F.relu( x )
		x = self.bn1(x)
		for i in range(self.nConv-1):
			x = self.conv2[i](x)
			x = F.relu( x )
			x = self.bn2[i](x)
		x = self.conv3(x)
		x = self.bn3(x)
		return x



import pandas as pd
import numpy as np
import cv2

def scan_contour(spots, scan_x=True, shape="hexagon"):
	#shape="hexagon" # For 10X Vsium, shape="square" for ST data
	if scan_x:
		array_a="array_x"
		array_b="array_y"
		pixel_a="pixel_x"
		pixel_b="pixel_y"
	else:
		array_a="array_y"
		array_b="array_x"
		pixel_a="pixel_y"
		pixel_b="pixel_x"
	upper, lower={}, {}
	uniq_array_a=sorted(set(spots[array_a]))
	if shape=="hexagon":
		for i in range(len(uniq_array_a)-1):
			a1=uniq_array_a[i]
			a2=uniq_array_a[i+1]
			group=spots.loc[spots[array_a].isin([a1, a2]),:]
			lower[np.min(group[pixel_a])]=np.min(group[pixel_b])
			upper[np.min(group[pixel_a])]=np.max(group[pixel_b])
			#print(a1, lower[np.min(group[pixel_a])], upper[np.min(group[pixel_a])])
		a1=uniq_array_a[-1]
		a2=uniq_array_a[-2]
		group=spots.loc[spots[array_a].isin([a1, a2]),:]
		lower[np.min(group[pixel_a])]=np.min(group[pixel_b])
		upper[np.min(group[pixel_a])]=np.max(group[pixel_b])
		#print(a1, lower[np.min(group[pixel_a])], upper[np.min(group[pixel_a])])
	elif shape=="square":
		for i in range(len(uniq_array_a)-1):
			a1=uniq_array_a[i]
			group=spots.loc[spots[array_a]==a1,:]
			lower[np.min(group[pixel_a])]=np.min(group[pixel_b])
			upper[np.min(group[pixel_a])]=np.max(group[pixel_b])
		a1=uniq_array_a[-1]
		group=spots.loc[spots[array_a]==a1,:]
		lower[np.min(group[pixel_a])]=np.min(group[pixel_b])
		upper[np.min(group[pixel_a])]=np.max(group[pixel_b])
	else:
		print("Error, unknown shape, pls specify 'square' or 'hexagon'.")
	lower=np.array(list(lower.items())[::-1]).astype("int32")
	upper=np.array(list(upper.items())).astype("int32")
	cnt=np.concatenate((upper, lower), axis=0)
	cnt=cnt.reshape(cnt.shape[0], 1, 2)
	if scan_x:
		cnt=cnt[:, : , [1, 0]]
	return cnt

def cv2_detect_contour(img, 
	CANNY_THRESH_1 = 100,
	CANNY_THRESH_2 = 200,
	apertureSize=5,
	L2gradient = True,
	all_cnt_info=False):
	if len(img.shape)==3:
		gray = cv2.cvtColor(img,cv2.COLOR_BGR2GRAY)
	elif len(img.shape)==2:
		gray=(img*((1, 255)[np.max(img)<=1])).astype(np.uint8)
	else:
		print("Image format error!")
	edges = cv2.Canny(gray, CANNY_THRESH_1, CANNY_THRESH_2,apertureSize = apertureSize, L2gradient = L2gradient)
	edges = cv2.dilate(edges, None)
	edges = cv2.erode(edges, None)
	cnt_info = []
	cnts, _ = cv2.findContours(edges, cv2.RETR_LIST, cv2.CHAIN_APPROX_NONE)
	for c in cnts:
		cnt_info.append((c,cv2.isContourConvex(c),cv2.contourArea(c),))
	cnt_info = sorted(cnt_info, key=lambda c: c[2], reverse=True)
	cnt=cnt_info[0][0]
	if all_cnt_info:
		return cnt_info
	else:
		return cnt


def cut_contour_boundary(cnt, x_min, x_max, y_min, y_max, enlarge):
	ret=cnt.copy()
	ret[:, : , 0][ret[:, : , 0]>y_max]=y_max
	ret[:, : , 0][ret[:, : , 0]<y_min]=y_min
	ret[:, : , 1][ret[:, : , 1]>x_max]=x_max
	ret[:, : , 1][ret[:, : , 1]<x_min]=x_min
	return ret

def scale_contour(cnt, scale):
	M = cv2.moments(cnt)
	cx = int(M['m10']/M['m00'])
	cy = int(M['m01']/M['m00'])
	cnt_norm = cnt - [cx, cy]
	cnt_scaled = cnt_norm * scale
	cnt_scaled = cnt_scaled + [cx, cy]
	cnt_scaled = cnt_scaled.astype(np.int32)
	return cnt_scaled











import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="TESLAforST", 
    version="1.2.0",
    author="Jian Hu",
    author_email="jianhu@pennmedicine.upenn.edu",
    description="TESLA: Deciphering tumor ecosystems at super-resolution from spatial transcriptomics",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/jianhuupenn/TESLA",
    packages=setuptools.find_packages(),
    install_requires=["torch","pandas","numpy","scipy","scanpy","anndata","sklearn", "numba"],
    #install_requires=[],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
)



import numpy as np
import pandas as pd
import time
import cv2
import argparse
import torch
import torch.nn as nn
import torch.nn.functional as F
import torch.optim as optim
import torch.nn.init
from torch.autograd import Variable
from . models import Spatial_CNN

class TESLA(object):
	def __init__(self):
		super(TESLA, self).__init__()

	def train(self, input,
		use_cuda=False, 
		train_refine=True, 
		radius=3, 
		nChannel=100, 
		lr=0.1, 
		minLabels=20, 
		maxIter=30, 
		stepsize_sim=1, 
		stepsize_con=10, 
		threshold=1000,
		plot_intermedium=False,
		plot_dir="./"):
		self.use_cuda=use_cuda 
		self.train_refine=train_refine 
		self.radius=radius 
		self.nChannel=nChannel 
		self.lr=lr 
		self.minLabels=minLabels 
		self.maxIter=maxIter 
		self.stepsize_sim=stepsize_sim 
		self.stepsize_con=stepsize_con 
		self.threshold=threshold
		self.plot_intermedium=plot_intermedium
		self.resize_height=input.shape[0]
		self.resize_width=input.shape[1]
		input = torch.from_numpy( np.array([input.transpose( (2, 0, 1) ).astype('float32')]) )
		if use_cuda:
			input = input.cuda()
		input = Variable(input)
		#--------------------------------------- Train ---------------------------------------
		self.model = Spatial_CNN(input.size(1), nConv=2, nChannel=self.nChannel, kernel_size_list=[5, 5],stride_list=[1, 1], padding_list=[2, 2])
		if use_cuda:
			self.model.cuda()
		# Similarity loss definition
		loss_fn = torch.nn.CrossEntropyLoss()
		# Continuity loss definition
		loss_hpy = torch.nn.L1Loss(size_average = True)
		loss_hpz = torch.nn.L1Loss(size_average = True)
		HPy_target = torch.zeros(self.resize_height-1, self.resize_width, nChannel)
		HPz_target = torch.zeros(self.resize_height, self.resize_width-1, nChannel)
		if use_cuda:
			HPy_target = HPy_target.cuda()
			HPz_target = HPz_target.cuda()
		optimizer = optim.SGD(self.model.parameters(), lr=lr, momentum=0.9)
		label_colours = np.random.randint(255,size=(100,3))
		start_time = time.time()
		self.model.train()
		for batch_idx in range(maxIter):
			# forwarding
			optimizer.zero_grad()
			output = self.model( input )[ 0 ]
			output = output.permute( 1, 2, 0 ).contiguous().view( -1, nChannel )
			outputHP = output.reshape( (self.resize_height, self.resize_width, nChannel) )
			HPy = outputHP[1:, :, :] - outputHP[0:-1, :, :]
			HPz = outputHP[:, 1:, :] - outputHP[:, 0:-1, :]
			lhpy = loss_hpy(HPy,HPy_target)
			lhpz = loss_hpz(HPz,HPz_target)
			_, target = torch.max( output, 1 )
			img_target = target.data.cpu().numpy()
			# Total number of clusters
			nLabels = len(np.unique(img_target))
			# Number of main clusters
			mainLabels=(pd.Series(img_target).value_counts()>=threshold).sum()
			#--------------------------Refine during training----------------------------------------
			if train_refine:
				pixel_num=pd.Series(img_target).value_counts()
				main_clusters=pixel_num.index[pixel_num>=threshold].tolist()
				minor_clusters=pixel_num.index[pixel_num<threshold].tolist()
				b_refine = img_target.reshape( (self.resize_height, self.resize_width))
				max_x, max_y=self.resize_width, self.resize_height
				replace_map={}
				for i in minor_clusters:
					nbs=[]
					xy=np.where(b_refine==i)
					for j in range(len(xy[0])):
						x, y=xy[0][j], xy[1][j]
						nbs=nbs+b_refine[max(0,x-radius):min(max_x,x+radius+1),max(0,y-radius):min(max_y,y+radius+1)].flatten().tolist()
						nbs_num=pd.Series(nbs).value_counts()
						if nbs_num.index[nbs_num.index.isin(main_clusters)].shape[0]>0:
							replace_map[i]=nbs_num.index[nbs_num.index.isin(main_clusters)][0]
				target=torch.from_numpy(pd.Series(img_target).replace(replace_map).to_numpy())
			#------------------------------------------------------------------
			loss1 = stepsize_sim * loss_fn(output, target)
			loss2 = stepsize_con * (lhpy + lhpz)
			loss=loss1+loss2
			loss.backward()
			optimizer.step()
			print (batch_idx, '/', maxIter, '|', ' label num :', nLabels, '|',' main clusters :',mainLabels,' | feature loss :', loss1.item(),' | spatial loss :', loss2.item())
			print("--- %s seconds ---" % (time.time() - start_time))
			if batch_idx%2==0:
				output = self.model(input)[0]
				output = output.permute(1, 2, 0).contiguous().view(-1, nChannel)
				_, target = torch.max(output, 1)
				img_target = target.data.cpu().numpy()
				img_target_rgb = np.array([label_colours[ c %  label_colours.shape[0]] for c in img_target])
				img_target_rgb = img_target_rgb.reshape(self.resize_height, self.resize_width, 3).astype( np.uint8 )
				if self.plot_intermedium:
					cv2.imwrite(plot_dir+str(batch_idx)+"_nL"+str(nLabels)+"_mL"+str(mainLabels)+".png", img_target_rgb)
			if mainLabels <= minLabels:
				print ("mainLabels", mainLabels, "reached minLabels", minLabels, ".")
				break

	def predict(self, input):
		input = torch.from_numpy( np.array([input.transpose( (2, 0, 1) ).astype('float32')]) )
		output = self.model( input )[ 0 ]
		output = output.permute( 1, 2, 0 ).contiguous().view( -1, self.nChannel )
		prob=output.data.cpu().numpy()
		prob=np.divide(np.exp(prob), np.sum(np.exp(prob), 1).reshape(-1, 1))
		_, pred = torch.max( output, 1 )
		pred = pred.data.cpu().numpy()
		return prob, pred











#Get connected components in an undirected graph
import pandas as pd
import numpy as np
class Graph:
	# init function to declare class variables
	def __init__(self, V, adj):
		self.V = V
		self.adj = adj
	def DFSUtil(self, v, tmp, visited):
		# Mark the current vertex as visited
		visited[v] = 1
		# Store the vertex to list
		tmp.append(v)
		# Repeat for all vertices adjacent to this vertex v
		nbr=self.adj[v][self.adj[v]==1].index.tolist()
		for i in nbr:
			if visited[i] == 0:
				tmp = self.DFSUtil(i, tmp, visited)
		return tmp
	def ConnectedComponents(self):
		visited = pd.Series([0]* len(self.V), index=self.V)
		cc = []
		for v in self.V:
			if visited[v] == 0:
				tmp = []
				cc.append(self.DFSUtil(v, tmp, visited))
		return cc
